@@ -1,8 +1,10 @@
 //! Parses `.obj` format which stores 3D mesh data
 
+use std::collections::hash_map::Entry;
 use std::collections::HashMap;
+use std::hash::Hash;
 use std::io::BufRead;
-use vec_map::VecMap;
+use std::mem;
 
 use crate::error::ObjResult;
 use crate::raw::lexer::lex;
@@ -213,7 +215,7 @@ pub fn parse_obj<T: BufRead>(input: T) -> ObjResult<RawObj> {
 
             // Grouping
             "g" => match args {
-                [name] => group_builder.try_start((*name).to_string())?,
+                [name] => group_builder.start((*name).to_string()),
                 _ => make_error!(
                     WrongNumberOfArguments,
                     "Expected group name parameter, but nothing has been supplied"
@@ -221,12 +223,12 @@ pub fn parse_obj<T: BufRead>(input: T) -> ObjResult<RawObj> {
             },
             "s" => match args {
                 ["off"] | ["0"] => smoothing_builder.end(),
-                [param] => smoothing_builder.try_start(param.parse()?)?,
+                [param] => smoothing_builder.start(param.parse()?),
                 _ => make_error!(WrongNumberOfArguments, "Expected only 1 argument"),
             },
             "mg" => match args {
                 ["off"] | ["0"] => merging_builder.end(),
-                [param] => merging_builder.try_start(param.parse()?)?,
+                [param] => merging_builder.start(param.parse()?),
                 _ => make_error!(WrongNumberOfArguments, "Expected only 1 argument"),
             },
             "o" => {
@@ -243,7 +245,7 @@ pub fn parse_obj<T: BufRead>(input: T) -> ObjResult<RawObj> {
             "d_interp" => unimplemented!(),
             "lod" => unimplemented!(),
             "usemtl" => match args {
-                [material] => mesh_builder.try_start((*material).to_string())?,
+                [material] => mesh_builder.start((*material).to_string()),
                 _ => make_error!(WrongNumberOfArguments, "Expected only 1 argument"),
             },
             "mtllib" => {
@@ -329,31 +331,28 @@ impl Counter {
 
 /// Helper for creating `groups`, `meshes`, `smoothing_groups` and `merging_groups` member of
 /// `Obj`.
-struct GroupBuilder<'a, T, K> {
+struct GroupBuilder<'a, K> {
     counter: &'a Counter,
-    current: Option<K>, // Some(K) if some group has been started
-    // None    otherwise
-    result: T,
+    /// `Some(K)` if some group has been started, `None` otherwise.
+    current: Option<K>,
+    result: HashMap<K, Group>,
 }
 
-impl<'a, T, K> GroupBuilder<'a, T, K>
+impl<'a, K> GroupBuilder<'a, K>
 where
-    T: Map<K, Group>,
-    K: Clone + Key,
+    K: Clone + Eq + Hash,
 {
     fn new(counter: &'a Counter) -> Self {
         GroupBuilder {
             counter,
             current: None,
-            result: T::new(),
+            result: HashMap::new(),
         }
     }
 
     fn with_default(counter: &'a Counter, default: K) -> Self {
-        let mut result = T::with_capacity(1);
-        result
-            .try_insert(default.clone(), Group::new((0, 0, 0)))
-            .unwrap();
+        let mut result = HashMap::with_capacity(1);
+        result.insert(default.clone(), Group::new((0, 0, 0)));
 
         GroupBuilder {
             counter,
@@ -363,43 +362,57 @@ where
     }
 
     /// Starts a group whose name is `input`.
-    fn try_start(&mut self, input: K) -> ObjResult<()> {
+    fn start(&mut self, input: K) {
         let count = self.counter.get();
-        if let Some(ref current) = self.current {
-            if *current == input {
-                return Ok(());
-            }
-            if self.result.get_mut(current).unwrap().end(count) {
-                let res = self.result.remove(&current);
-                assert!(res.is_some());
+
+        match self.current {
+            // Started same group twice, do nothing
+            Some(ref current) if *current == input => return,
+            // Changing group
+            Some(ref mut current) => {
+                // Close the past group and start a new one
+                let past = mem::replace(current, input.clone());
+                match self.result.entry(past) {
+                    Entry::Vacant(_) => unreachable!(),
+                    Entry::Occupied(mut e) => {
+                        let was_empty_group = e.get_mut().end(count);
+                        // Remove the past group if the past group is empty
+                        if was_empty_group {
+                            e.remove();
+                        }
+                    }
+                }
             }
+            // Starting a new group
+            None => self.current = Some(input.clone()),
         }
-        if let Some(ref mut group) = self.result.get_mut(&input) {
-            group.start(count);
-        } else {
-            let res = self.result.try_insert(input.clone(), Group::new(count))?;
-            assert!(res.is_none());
-        }
-        self.current = Some(input);
-        Ok(())
+
+        self.result
+            .entry(input)
+            .and_modify(|e| e.start(count))
+            .or_insert_with(|| Group::new(count));
     }
 
     /// Ends a current group.
     fn end(&mut self) {
-        if let Some(ref current) = self.current {
-            if self
-                .result
-                .get_mut(current)
-                .unwrap()
-                .end(self.counter.get())
-            {
-                let result = self.result.remove(current);
-                assert!(result.is_some());
+        match mem::replace(&mut self.current, None) {
+            // Closed group twice, do nothing
+            None => {}
+            // Closing a group
+            Some(current) => {
+                match self.result.entry(current) {
+                    Entry::Vacant(_) => unreachable!(),
+                    Entry::Occupied(mut e) => {
+                        let count = self.counter.get();
+                        let was_empty_group = e.get_mut().end(count);
+                        // Remove the past group if the past group is empty
+                        if was_empty_group {
+                            e.remove();
+                        }
+                    }
+                }
             }
-        } else {
-            return;
         }
-        self.current = None;
     }
 }
 
@@ -452,70 +465,6 @@ impl Group {
     }
 }
 
-/// Custom trait to interface `HashMap` and `VecMap`.
-trait Map<K: Key, V> {
-    fn new() -> Self;
-    fn with_capacity(capacity: usize) -> Self;
-    /// Interface of `insert` function.
-    fn try_insert(&mut self, _: K, _: V) -> ObjResult<Option<V>>;
-    /// Interface of `get_mut` function.
-    fn get_mut(&mut self, k: &K) -> Option<&mut V>;
-    /// Interface of `remove` function.
-    fn remove(&mut self, k: &K) -> Option<V>;
-}
-
-impl<V> Map<String, V> for HashMap<String, V> {
-    fn new() -> Self {
-        HashMap::new()
-    }
-    fn with_capacity(capacity: usize) -> Self {
-        HashMap::with_capacity(capacity)
-    }
-    fn try_insert(&mut self, k: String, v: V) -> ObjResult<Option<V>> {
-        Ok(self.insert(k, v))
-    }
-    fn get_mut(&mut self, k: &String) -> Option<&mut V> {
-        self.get_mut(k)
-    }
-    fn remove(&mut self, k: &String) -> Option<V> {
-        self.remove(k)
-    }
-}
-
-impl<V> Map<usize, V> for VecMap<V> {
-    fn new() -> Self {
-        VecMap::new()
-    }
-    fn with_capacity(capacity: usize) -> Self {
-        VecMap::with_capacity(capacity)
-    }
-    fn try_insert(&mut self, k: usize, v: V) -> ObjResult<Option<V>> {
-        use crate::error::{LoadError, LoadErrorKind, ObjError};
-
-        // Biggest group number will be 67108864, which is large enough.
-        // It will consume 64MiB of memory in the worst case.
-        if k > 64 * (1 << 20) {
-            return Err(ObjError::Load(LoadError::new(
-                LoadErrorKind::TooBigGroupNumber,
-                "Currently biggest group number allowed is 67108864.",
-            )));
-        }
-        Ok(self.insert(k, v))
-    }
-    fn get_mut(&mut self, k: &usize) -> Option<&mut V> {
-        self.get_mut(*k)
-    }
-    fn remove(&mut self, k: &usize) -> Option<V> {
-        self.remove(*k)
-    }
-}
-
-/// A trait which should be implemented by a type passed into `Key` of `Map`.
-trait Key: Eq {}
-
-impl Key for String {}
-impl Key for usize {}
-
 /// Low-level Rust binding for `.obj` format.
 pub struct RawObj {
     /// Name of the object.
@@ -544,9 +493,9 @@ pub struct RawObj {
     /// Geometries which consist in a same material.
     pub meshes: HashMap<String, Group>,
     /// Smoothing groups.
-    pub smoothing_groups: VecMap<Group>,
+    pub smoothing_groups: HashMap<usize, Group>,
     /// Merging groups.
-    pub merging_groups: VecMap<Group>,
+    pub merging_groups: HashMap<usize, Group>,
 }
 
 /// The `Point` type which stores the index of the position vector.
